from __future__ import absolute_import, unicode_literals
import os
from celery import Celery
from fastapi_pundra.common.scheduler.schedule import bind_beat_schedule
from importlib import import_module
from dotenv import load_dotenv

load_dotenv()

def create_celery_app(project_name: str, broker_type: str = 'redis'):
  # Get project base path from environment
  project_base_path = os.getenv('PROJECT_BASE_PATH', 'app')
  
  # Dynamically import schedules based on PROJECT_BASE_PATH
  schedules_module = import_module(f'{project_base_path}.config.scheduler')
  schedules = schedules_module.schedules

  app = Celery(project_name)
  app.conf.update(
    broker_url=os.getenv('CELERY_BROKER_URL'),
    result_backend=os.getenv('CELERY_RESULT_BACKEND'),
    timezone='UTC',
    enable_utc=True,
  )

  if broker_type == 'redis':
    app.conf.beat_scheduler = 'redbeat.RedBeatScheduler'
    app.conf.redbeat_redis_url = os.getenv('CELERY_BROKER_URL')

  # Define default task modules that should always be included
  default_task_modules = [
    'fastapi_pundra.common.mailer.task'
  ]
  
  # Define project-specific task modules
  project_task_modules = [
    f"{project_base_path}.tasks"
  ]
  
  # Combine all task modules
  all_task_modules = project_task_modules + default_task_modules
  app.autodiscover_tasks(all_task_modules)

  app.conf.beat_schedule = bind_beat_schedule(schedules=schedules) 

  return app


from graphql import GraphQLError

class NotFoundError(GraphQLError):
    def __init__(self, message: str):
        super().__init__(
            message=message,
            extensions={
                "code": "NOT_FOUND",
                "category": "NotFound",
            }
        )


class DtoValidationError(GraphQLError):
    @staticmethod
    def format_validation_errors(validation_errors) -> dict:
        errors = {}
        for error in validation_errors:
            field = error['loc'][0]
            message = f"{field} {error['msg']}"
            if field not in errors:
                errors[field] = []
            errors[field].append(message)
        return errors

    def __init__(self, message: str, validation_errors=None):
        extensions = {
            "code": "VALIDATION_ERROR",
            "category": "Validation",
        }
        
        if validation_errors:
            formatted_errors = self.format_validation_errors(validation_errors)
            extensions["validation_errors"] = formatted_errors

        super().__init__(
            message=message,
            extensions=extensions
        )

class BadRequestError(GraphQLError):
    def __init__(self, message: str):
        super().__init__(
            message=message,
            extensions={
                "code": "BAD_REQUEST",
                "category": "BadRequest",
            }
        )

class ForbiddenError(GraphQLError):
    def __init__(self, message: str):
        super().__init__(
            message=message,
            extensions={
                "code": "FORBIDDEN",
                "category": "Forbidden",
            }
        )

class ConflictError(GraphQLError):
    def __init__(self, message: str):
        super().__init__(
            message=message,
            extensions={
                "code": "CONFLICT",
                "category": "Conflict",
            }
        )

class DuplicateError(GraphQLError):
    def __init__(self, message: str):
        super().__init__(
            message=message,
            extensions={
                "code": "DUPLICATE",
                "category": "Duplicate",
            }
        )

class InternalServerError(GraphQLError):
    def __init__(self, message: str = "An unexpected error occurred"):
        super().__init__(
            message=message,
            extensions={
                "code": "INTERNAL_SERVER_ERROR",
                "category": "ServerError",
            }
        )


class RateLimitError(GraphQLError):
    def __init__(self, message: str, remaining_time: int = None):
        extensions = {
            "code": "RATE_LIMIT_EXCEEDED",
            "category": "RateLimit",
        }
            
        super().__init__(
            message=message,
            extensions=extensions
        )


from typing import Any, Dict, Optional

class BaseAPIException(Exception):
    """Base exception for all API exceptions"""
    def __init__(
        self,
        message: str = "An error occurred",
        status_code: int = 500,
        errors: Optional[Dict[str, Any]] = None
    ):
        self.message = message
        self.status_code = status_code
        self.errors = errors or {}
        super().__init__(self.message)

    def to_dict(self) -> Dict[str, Any]:
        output = {
            'success': False,
            'message': self.message
        }
        if self.errors:
            output['errors'] = self.errors
        return output

class ValidationException(BaseAPIException):
    """Validation error exception"""
    def __init__(self, errors: Dict[str, Any]):
        super().__init__(
            message="Validation error",
            status_code=422,
            errors=errors
        )

class NotFoundException(BaseAPIException):
    """Resource not found exception"""
    def __init__(self, message: str = "Resource not found"):
        super().__init__(
            message=message,
            status_code=404
        )

class ItemNotFoundException(BaseAPIException):
    """Item not found exception"""
    def __init__(self, message: str = "Item not found"):
        super().__init__(
            message=message,
            status_code=404
        )

class UnauthorizedException(BaseAPIException):
    """Unauthorized access exception"""
    def __init__(self, message: str = "Unauthorized access"):
        super().__init__(
            message=message,
            status_code=401
        )

class ForbiddenException(BaseAPIException):
    """Forbidden access exception"""
    def __init__(self, message: str = "Forbidden access"):
        super().__init__(
            message=message,
            status_code=403
        )

class BadRequestException(BaseAPIException):
    """Bad request exception"""
    def __init__(self, message: str = "Bad request", errors: Optional[Dict[str, Any]] = None):
        super().__init__(
            message=message,
            status_code=400,
            errors=errors
        )

class ConflictException(BaseAPIException):
    """Conflict exception"""
    def __init__(self, message: str = "Resource conflict"):
        super().__init__(
            message=message,
            status_code=409
        )

class MethodNotAllowedException(BaseAPIException):
    """Method not allowed exception"""
    def __init__(self, message: str = "Method not allowed"):
        super().__init__(
            message=message,
            status_code=405
        )


import functools
from typing import Type
import pydantic
from fastapi_pundra.gql_berry.exception import DtoValidationError

def dto_validation(dto_class: Type[pydantic.BaseModel]):
    def decorator(func):
        @functools.wraps(func)
        def wrapper(*args, **kwargs):
            try:
                # Find the input data argument (usually the last one)
                input_data = next((v for k, v in kwargs.items() if hasattr(v, '__dict__')), None)
                if input_data:
                    validated_data = dto_class(**input_data.__dict__)
                    # Replace the input data with validated data
                    kwargs = {k: validated_data if v == input_data else v for k, v in kwargs.items()}
                return func(*args, **kwargs)
            except pydantic.ValidationError as e:
                raise DtoValidationError(
                    message="Validation error",
                    validation_errors=e.errors()
                )
        return wrapper
    return decorator 


import os
from bs4 import BeautifulSoup
from premailer import Premailer
from fastapi_pundra.common.helpers import base_path
from dotenv import load_dotenv

load_dotenv()

def inline_css(html_code):
    # Parse the HTML code
    soup = BeautifulSoup(html_code, 'html.parser')

    # Find all link tags with a "stylesheet" rel attribute
    link_tags = soup.find_all('link', rel='stylesheet')

    # Extract the href attribute values
    css_files = [link['href'] for link in link_tags]

    project_base_path = os.getenv('PROJECT_BASE_PATH', 'app')

    # Read the content of each CSS file and include it in the HTML
    mail_template_dir = os.path.join(base_path(), project_base_path, 'templates', 'mails')
    for css_file in css_files:
        css_file_path = os.path.join(mail_template_dir, css_file)
        # Check if the file exists
        if os.path.exists(css_file_path):
            with open(css_file_path, 'r') as file:
                css_content = file.read()
                # Create a new style tag with the CSS content
                style_tag = soup.new_tag('style')
                style_tag.string = css_content
                # Replace the link tag with the new style tag
                link_tag = soup.find('link', href=css_file)
                link_tag.replace_with(style_tag)
        else:
            print('file not found')
            pass
          
    for tag in link_tags:
        tag.extract()

    # Get the modified HTML as a string
    modified_html = str(soup)
    
    # ascii issue: https://github.com/peterbe/premailer/issues/249#issuecomment-1541119706
    the_html_with_ascii = modified_html.encode("ascii", "xmlcharrefreplace").decode("ascii")
    
    # intance of premailer.
    premailer_instance = Premailer(
        the_html_with_ascii,
        strip_important=False,
        remove_classes=True
      )
    
    processed_html = premailer_instance.transform()

    return processed_html


from .task import send_email_queue_task
from typing import List

async def send_mail_queue(
    subject: str, 
    to: List[str], 
    template_name: str, 
    context: dict, 
    cc: List[str] | str = None, 
    bcc: List[str] | str = None, 
    reply_to: List[str] | str = None, 
    delay: int = 30
):
    send_email_queue_task.apply_async(
        args=[subject, to, template_name, context, cc, bcc, reply_to],
        countdown=delay
    ) 


from typing import Any

def openapi_request_body_schema(
    schema_class: type,
    *,
    required: bool = True,
    description: str | None = None,
    content_type: str = "application/json",
    examples: dict[str, dict[str, Any]] | None = None,
) -> dict[str, Any]:
    """
    Generate OpenAPI request body configuration for FastAPI routes.

    Args:
        schema_class: Pydantic model class to use for the request body schema
        required: Whether the request body is required (default: True)
        description: Optional description for the request body
        content_type: Content type for the request body (default: "application/json")
        examples: Optional dictionary of example requests (name -> example object)

    Returns:
        Dictionary containing OpenAPI requestBody configuration
    """
    content_config: dict[str, Any] = {
        "schema": schema_class.model_json_schema()
    }

    if examples:
        content_config["examples"] = examples

    config: dict[str, Any] = {
        "requestBody": {
            "content": {content_type: content_config},
            "required": required,
        }
    }

    if description:
        config["requestBody"]["description"] = description

    return config



import importlib
from typing import List, Type
from pathlib import Path

def load_resolver_class(path: str) -> Type:
    """
    Dynamically load a resolver class from a given path.
    
    Args:
        path (str): The full path to the resolver class
        
    Returns:
        Type: The loaded resolver class
        
    Raises:
        ValueError: If the path is not properly formatted (must contain at least one dot)
    """
    if '.' not in path:
        raise ValueError(f"Invalid path format: {path}. Path must be in format 'module.class'")
    
    module_path, class_name = path.rsplit('.', 1)
    try:
        module = importlib.import_module(module_path)
        return getattr(module, class_name)
    except ImportError as e:
        raise ImportError(f"Could not import module '{module_path}': {str(e)}")
    except AttributeError as e:
        raise AttributeError(f"Could not find class '{class_name}' in module '{module_path}': {str(e)}")

def discover_resolvers(base_path: str = "app/graphql") -> List[str]:
    """
    Discover all resolver classes in the specified base path.
    Looks for files ending with '_query_resolvers.py' and '_mutation_resolvers.py'.
    
    Args:
        base_path (str): The base path to start searching for resolvers
        
    Returns:
        List[str]: List of fully qualified resolver class paths
    """
    query_resolver_paths = []
    mutation_resolver_paths = []
    base = Path(base_path)
    
    # Walk through all subdirectories in the graphql folder
    for path in base.rglob("*_query_resolvers.py"):
        try:
            # Find the position of 'app' in the path
            module_parts = path.parts
            app_index = module_parts.index('app')
            # Take the path from 'app' onwards
            module_path = '.'.join(module_parts[app_index:-1] + (module_parts[-1][:-3],))
            
            # Import the module using importlib
            module = importlib.import_module(module_path)
            
            # Look for classes that end with 'QueryResolver'
            for name, obj in module.__dict__.items():
                if isinstance(obj, type) and name.endswith('QueryResolvers'):
                    query_resolver_paths.append(f"{module_path}.{name}")
        except Exception as e:
            print(f"Error loading query resolver module {str(path)}: {str(e)}")
            continue

    # Walk through all subdirectories looking for mutation resolvers
    for path in base.rglob("*_mutation_resolvers.py"):
        try:
            module_parts = path.parts
            app_index = module_parts.index('app')
            module_path = '.'.join(module_parts[app_index:-1] + (module_parts[-1][:-3],))
            
            module = importlib.import_module(module_path)
            
            # Look for classes that end with 'MutationResolver'
            for name, obj in module.__dict__.items():
                if isinstance(obj, type) and name.endswith('MutationResolvers'):
                    mutation_resolver_paths.append(f"{module_path}.{name}")
        except Exception as e:
            print(f"Error loading mutation resolver module {str(path)}: {str(e)}")
            continue
    
    return [query_resolver_paths, mutation_resolver_paths]


import strawberry
from typing import List, Optional, TypeVar, Generic

T = TypeVar('T')

@strawberry.type(description="Generic list type with pagination")
class PaginatedList(Generic[T]):
    data: List[T]
    pagination: strawberry.scalars.JSON
    additional_data: Optional[strawberry.scalars.JSON] = None


from celery import shared_task
from typing import List
import asyncio
from .mail import send_mail

@shared_task(name="pundra_send_email_queue_task")
def send_email_queue_task(subject: str, to: List[str], template_name: str, context: dict, cc: List[str] | str = None, bcc: List[str] | str = None, reply_to: List[str] | str = None, logger = None):
    try:
        coro = send_mail(
            subject=subject,
            to=to,
            template_name=template_name,
            context=context,
            cc=cc,
            bcc=bcc,
            reply_to=reply_to
        )
        
        result = asyncio.run(coro)
        return "Email sent"
    except Exception as e:
        print(f"Error sending email: {str(e)}")
        raise


from setuptools import setup, find_packages
import os
import re
import shutil
import sys

def get_version(package):
    """
    Return package version as listed in `__version__` in `init.py`.
    """
    init_py = open(os.path.join(package, '__init__.py')).read()
    return re.search("__version__ = ['\"]([^'\"]+)['\"]", init_py).group(1)


version = get_version('fastapi_pundra')

if sys.argv[-1] == 'publish':
    os.system("python setup.py sdist bdist_wheel")
    if os.system("twine check dist/*"):
        print("twine check failed. Packages might be outdated.")
        print("Try using `pip install -U twine wheel`.\nExiting.")
        sys.exit()
    os.system("twine upload dist/*")
    print("You probably want to also tag the version now:")
    print("  git tag -a %s -m 'version %s'" % (version, version))
    print("  git push --tags")
    shutil.rmtree('dist')
    shutil.rmtree('build')
    shutil.rmtree('fastapi_pundra.egg-info')
    sys.exit()
    
def read(f):
    with open(f, 'r', encoding='utf-8') as file:
        return file.read()
    
setup(
    # Package metadata
    name='fastapi-pundra',
    version=version,
    url='https://github.com/code4mk/fastapi-pundra',
    author='Mostafa Kamal',
    author_email='hiremostafa@gmail.com',
    description='Pundra: Your FastAPI Companion for Productivity',
    long_description=read('README.md'),
    long_description_content_type='text/markdown',
    keywords=['python', 'fastapi', 'rest_api', 'pundra', 'strawberry', 'graphql', 'code4mk'],
    license='MIT',
    # Package configuration
    packages=find_packages(exclude=['tests*']),
    python_requires='>=3.6',
    
    # Dependencies
    install_requires=[
        "requests",
        "python-dotenv",
    ],

    # Classifiers
    classifiers=[
        'Development Status :: 3 - Alpha',
        'Intended Audience :: Developers',
        'License :: OSI Approved :: MIT License',
        'Framework :: FastAPI',
        'Programming Language :: Python :: 3.10',
        'Programming Language :: Python :: 3.11',
        'Programming Language :: Python :: 3.12',
        'Operating System :: OS Independent',
    ],
    project_urls={
        'Source': 'https://github.com/code4mk/fastapi-pundra',
        'Changelog': 'https://github.com/code4mk/fastapi-pundra/blob/main/CHANGELOG.md',
    },
)



from fastapi import Request
from sqlalchemy import desc
from pydantic import BaseModel

async def the_query(request: Request, name: str | None = None) -> dict[str, str] | str | None:
    """Get the query parameters from the request."""
    data = {}

    if request.query_params:
        data =  request.query_params
    elif request.headers.get("Content-Type") == "application/json":
        data = await request.json()
    else:
        data = await request.form()

    if name:
        return data.get(name)
    return data

def the_sorting(request: Request, query, default_sort=""):
    """Sort a SQLAlchemy query based on query parameters.
    
    Example:
        # For a request with URL: /api/items?sort=name,-created_at
        # This will sort by name (ascending) and created_at (descending)
    """
    sort_params = request.query_params.get("sort") or default_sort
    
    if sort_params:
        sort_fields = sort_params.split(",")
        ordering = []
        model_class = query.column_descriptions[0]['entity']
        
        # Get all valid column names from the model
        valid_columns = model_class.__table__.columns.keys()

        for field in sort_fields:
            field_name = field[1:] if field.startswith("-") else field
            
            # Check if the field exists in the model
            if field_name in valid_columns:
                if field.startswith("-"):
                    ordering.append(desc(getattr(model_class, field_name)))
                else:
                    ordering.append(getattr(model_class, field_name))
                
        if ordering:
            query = query.order_by(*ordering)
        
    return query

def get_serialize_data(schema: BaseModel, data: dict) -> dict:
    """Get the serialized data."""
    return schema.model_validate(data).model_dump()


from datetime import datetime, timedelta, UTC
from jose import jwt
import os

# Default values for development, but should be overridden in production
SECRET_KEY = os.getenv("SECRET_KEY", "my-secret-key")
ALGORITHM = os.getenv("ALGORITHM", "HS256")
ACCESS_TOKEN_EXPIRE_MINUTES = 525600  # 1 year (365 days * 24 hours * 60 minutes)
REFRESH_TOKEN_EXPIRE_DAYS = 730      # 2 years

def create_access_token(data: dict, expire_minutes: int = ACCESS_TOKEN_EXPIRE_MINUTES):
    to_encode = data.copy()
    expire = datetime.now(UTC) + timedelta(minutes=expire_minutes)
    to_encode.update({"exp": expire})
    return jwt.encode(to_encode, SECRET_KEY, algorithm=ALGORITHM)

def create_refresh_token(data: dict, expire_days: int = REFRESH_TOKEN_EXPIRE_DAYS):
    to_encode = data.copy()
    expire = datetime.now(UTC) + timedelta(days=expire_days)
    to_encode.update({"exp": expire})
    return jwt.encode(to_encode, SECRET_KEY, algorithm=ALGORITHM)

def decode_token(token: str) -> dict:
    """
    Decode and verify a JWT token
    Returns the decoded payload if valid, raises JWTError if invalid
    """
    if not token:
        raise jwt.JWTError("Token is missing")
        
    try:
        decoded_token = jwt.decode(token, SECRET_KEY, algorithms=[ALGORITHM])
        return decoded_token
    except jwt.JWTError as e:
        raise jwt.JWTError(f"Invalid token: {str(e)}")


from fastapi import FastAPI, Request
from fastapi.responses import JSONResponse
from fastapi_pundra.rest.exceptions import BaseAPIException, MethodNotAllowedException
from datetime import datetime

def setup_exception_handlers(app: FastAPI):
    @app.exception_handler(BaseAPIException)
    async def api_exception_handler(request: Request, exc: BaseAPIException):
        error_response = exc.to_dict()
        error_response["path"] = request.url.path
        error_response["type"] = exc.__class__.__name__
        error_response["timestamp"] = datetime.now().isoformat()

        return JSONResponse(
            status_code=exc.status_code,
            content=error_response
        )
    
    @app.middleware("http")
    async def exception_handling(request: Request, call_next):
        try:
            response = await call_next(request)
            
            # Method not allowed exception.
            if response.status_code == 405:
                raise MethodNotAllowedException(message="Method not allowed")
            
            return response
        except Exception as exc:
            error_details = {
                "success": False,
                "message": str(exc),
                "type": exc.__class__.__name__,
                "path": request.url.path,
                "timestamp": datetime.now().isoformat()
            }

            return JSONResponse(status_code=500, content=error_details)